/**************************************************************************************************************************
// This script was authored by Lorne Brooks - Email: lbtools2k@gmail.com
//
//
// REVISION HISTORY:
// Version 1.4 03/05/20:
//						- UI now remembers its last screen position.
//						- Added a toolbar icon.
//						- Converted dialog to a floater with rollouts.
//						- Added simple bake methods separate from the legacy advanced method. Can now bake just the current motion, or to
//						  a picked target with optional alignment to that target object.
//						- Fixed issues with bake advanced link to target and unlink from current parent.
//						- General script cleanup & improvements.
// Version 1.31 17/11/16 (hotfix):
//						- Fixed the baking time range settings to update correctly if the range is adjusted in the scene while the UI is open.
// Version 1.3 16/11/16;
//						- Major UI changes; removed redundant buttons and added more features and options.
//						- Consolodated previously separate bake buttons into one.
//						- Baking now resets all transform controllers on selected objects to their defaults (PRS, PositionXYZ, RotationXYZ, Scale Bezier).
//						  This effectively removes any type of constraint, as well as animation layers.
//						- Added the ability to bake the motion from the current parents of each selected object.
//						- Added the ability to align to either the current parent, a new parent (shared), or a target object.
//						- Added a fake undo button (basically it uses a scene hold during the bake process then a fetch when undoing).
// Version 1.2 27/10/14;
//						- Email change for help contact.
// Version 1.1 11/09/14;
//						- Recoded script as a struct to minimize global variables.
//						- Major UI changes.
//						- Bug fixes.
// Version 1.0 16/05/08;
//						- First public release.
//
//
// This script is not copyrighted, please feel free to modify or distribute.
//
**************************************************************************************************************************/

global g_LBTools_BO

struct s_LBTools_bakeObjects
(

/*==================================================================================
// Structure variables
====================================================================================*/

bakeObjectsFloater,
timeRange_Rollout,
bakeObjectsSimple_Rollout, --// Simple no-frills baking
bakeObjects_Rollout, --// More advanced baking (legacy)
aboutScript_rollout,
f_openBakeObjectsUI,
bakeObjectsINI = getDir #plugcfg + @"\LBTools_BakeObjects.ini",
dPos,
animStart = animationrange.start,
animEnd = animationrange.end,
stepSize = 1,

/*==================================================================================
// Structure functions
/*==================================================================================
	
f_validateObject
f_removeBindings
f_unlockAllTracks
f_bakeObjects

/*==================================================================================*/

fn f_validateObject obj =
(
	local isValid = false
	if obj != undefined AND isValidNode obj AND NOT isDeleted obj then
	(
		isValid = true
	)
	isValid
),

/*----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------*/

fn f_removeBindings obj =
(
	obj.transform.controller = PRS()
	obj.position.controller = Position_XYZ()
	obj.rotation.controller = Euler_XYZ()
	obj.scale.controller = Bezier_Scale()
	obj.parent = undefined
),

/*----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------*/

fn f_unlockAllTracks obj =
(
	LockedTracksMan.SetLocks false #(obj[3][1], obj[3][2], obj[3][3]) #(obj[3], obj[3], obj[3]) #(1,2,3) true		
),

/*----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------*/

fn f_bakeObjects =
(
	rollout timeRange_Rollout "Time Range"
	(
		group ""
		(
			radiobuttons rdobtns_timeRange labels:#("Active Range", "Set Range") columns:2 align:#center offset:[10,0] default:1
			spinner spn_startTime "Start: " range:[0,1000000, g_LBTools_BO.animStart] type:#integer fieldwidth:40 align:#left enabled:false across:2
			spinner spn_endTime "End: " range:[1,1000000, g_LBTools_BO.animEnd] type:#integer fieldwidth:40 align:#center enabled:false
			spinner spn_stepSize "Step: " range:[1,1000000, g_LBTools_BO.stepSize] type:#integer fieldwidth:40 align:#center
		)

		on timeRange_Rollout moved pos do
		(
			g_LBTools_BO.dPos = pos
			setIniSetting g_LBTools_BO.bakeObjectsINI "Bake Objects" "Position" (pos as string)
		)
		
		on rdobtns_timeRange changed state do
		(
			case rdobtns_timeRange.state of
			(
				1: (
						spn_startTime.enabled = false
						spn_endTime.enabled = false
					)
				2: (
						spn_startTime.enabled = true
						spn_endTime.enabled = true
					)
			)
		)

		on spn_startTime changed val do
		(
			if val >= spn_endTime.value do
			(
				spn_endTime.value = val + 1
			)
			g_LBTools_BO.animStart = val
		)
		
		on spn_endTime changed val do
		(
			if val <= spn_startTime.value do
			(
				spn_startTime.value = val - 1
			)
			g_LBTools_BO.animEnd = val
		)
		
		on spn_stepSize changed val do
		(
			g_LBTools_BO.stepSize = val
		)
	) --// end rollout
	
	rollout bakeObjectsSimple_Rollout "Bake Simple"
	(
		local targetObj
		
		label lbl_esc1 "[ hold ESC to cancel ]" align:#center
		group ""
		(
			button btn_simpleBake ">>  Bake Current Motion  <<" align:#center width:190 height:30 tooltip:"Bake the selected objects based on their current motion"
		)
		group ""
		(
			button btn_selectSimpleTarget "|" width:10 height:25 across:4 align:#left offset:[-5,0] tooltip:"Select the target object (if available)"
			label lbl_simpleTarget "Target:" across:3 align:#left offset:[-37,5]
			pickbutton pbtn_simpleTarget "Pick a TARGET" autoDisplay:true align:#center offset:[-10,0] tooltip:"Pick a target object to bake its motion onto the selected objects"
			button btn_clearSimpleTarget "X" align:#right offset:[7,0] width:20 hight:25 tooltip:"Clear the target object"
			button btn_simpleBakeToTarget ">>  Bake To Target  Motion <<" align:#center enabled:false width:190 height:30 tooltip:"Bake the selected objects to match the motion of a picked target object"
			checkbox chkbx_simpleAlign "Align To Target" checked:false enabled:false tooltip:"Align the selected objects to the picket target object after baking"
		)
		group ""
		(
			button btn_undoSimpleBake "-------  UNDO BAKE  -------" width:190 height:20 align:#center tooltip:"The bake process will first HOLD the scene file. Undo Bake will FETCH the scene file back"
		)
		
		on btn_undoSimpleBake pressed do
		(
			if (queryBox "Are you sure you want to UNDO BAKE?" title:"") do
			(
				fetchMaxFile quiet:true
				g_LBTools_BO.f_openBakeObjectsUI()
			)
		)
		
		on btn_simpleBake pressed do
		(
			if selection.count > 0 then
			(
				holdMaxFile()
				setCommandPanelTaskMode #create
				local a_selObjs = selection as array
				for obj in a_selObjs while not keyboard.escPressed do
				(
					local copyObj = copy obj
					g_LBTools_BO.f_unlockAllTracks obj
					g_LBTools_BO.f_removeBindings obj
					with animate on
					(
						for t = g_LBTools_BO.animStart to g_LBTools_BO.animEnd by g_LBTools_BO.stepSize do
						(
							at time t (obj.transform = copyObj.transform)
						)
					)
					delete copyObj
				)
			)
			else
			(
				messageBox "Error: Please select one or more objects first." title:""
			)
		)
		
		on btn_selectSimpleTarget pressed do
		(
			if g_LBTools_BO.f_validateObject targetObj then
			(
				select targetObj
			)
			else
			(
				messageBox "Error: The TARGET object is not valid.\n\nPlease check to see if it exists and pick it again."
				g_LBTools_BO.f_openBakeObjectsUI()
			)
		)
		
		on btn_clearSimpleTarget pressed do
		(
			pbtn_simpleTarget.object = targetObj = undefined
			btn_simpleBakeToTarget.enabled = false
			chkbx_simpleAlign.enabled = false
		)
		
		on pbtn_simpleTarget picked obj do
		(
			pbtn_simpleTarget.object = targetObj = obj
			btn_simpleBakeToTarget.enabled = true
			chkbx_simpleAlign.enabled = true
		)
		
		on btn_simpleBakeToTarget pressed do
		(
			if g_LBTools_BO.f_validateObject targetObj then
			(
				if selection.count > 0 then
				(
					holdMaxFile()
					setCommandPanelTaskMode #create
					local a_selObjs = selection as array
					for obj in a_selObjs while not keyboard.escPressed do
					(
						local copyTransform = obj.transform
						g_LBTools_BO.f_unlockAllTracks obj
						g_LBTools_BO.f_removeBindings obj
						with animate on
						(
							for t = g_LBTools_BO.animStart to g_LBTools_BO.animEnd by g_LBTools_BO.stepSize do
							(
								at time t (obj.transform = targetObj.transform)
							)
						)
						if chkbx_simpleAlign.checked == false then
						(
							obj.transform = copyTransform
						)
					)
				)
				else
				(
					messageBox "Error: Please select one or more objects first." title:""
				)
			)
			else
			(
				messageBox "Error: The TARGET object is not valid.\n\nPlease check to see if it exists and pick it again."
				g_LBTools_BO.f_openBakeObjectsUI()
			)
		)
	) --// end rollout
	
	rollout bakeObjects_Rollout "Bake Advanced"
	(
		local newParentObj
		local targetObj
		
		group ""
		(
			button btn_selectParent "|" width:10 height:25 across:4 align:#left offset:[-5,0] tooltip:"Select the new parent object (if available)"
			label lbl_parent "Parent:" align:#left offset:[-37,5]
			pickButton pbtn_parent "Pick a NEW PARENT" autoDisplay:true align:#center offset:[-10,0] tooltip:"Pick another object to become the shared new parent of the selected objects" 
			button btn_clearParent "X" align:#right offset:[7,0] width:20 hight:25 tooltip:"Clear the new parent object"
			button btn_selectTarget "|" width:10 height:25 across:4 align:#left offset:[-5,0] tooltip:"Select the target object (if available)"
			label lbl_target "Target:" across:3 align:#left offset:[-37,5]
			pickbutton pbtn_target "Pick a TARGET" autoDisplay:true align:#center offset:[-10,0] tooltip:"Pick a target object to bake its motion onto the selected objects"
			button btn_clearTarget "X" align:#right offset:[7,0] width:20 hight:25 tooltip:"Clear the target object"
			checkbox chkbx_linkToTarget "Link To Target Object" enabled:false align:#left offset:[15,0] tooltip:"Link selected objects to the picked target object"
			checkbox chkbx_unlink "Unlink From Current Parent" enabled:false checked:true align:#left offset:[15,0] tooltip:"Unlink selected objects from their current parent"
			checkbox chkbx_currentParent "Bake From Current Parent" align:#left tooltip:"Overrides a picked target to bake the motion of the current parent (of each selected object) onto the selected objects"
			button btn_bakeAdvanced ">>  Bake Advanced  <<" align:#center width:190 height:30 tooltip:"Bake the selected objects using the advanced settings"
			label lbl_esc2 "[ hold ESC to cancel ]" align:#center
			radiobuttons rdobtns_align labels:#("Parent", "New Parent", "Target") tooltip:#("Align to current parent object","Align to the new parent object","Align to the target object") default:1 columns:3 enabled:false align:#center offsets:#([30,0], [5,0], [0,0])
			checkbox chkbx_align "Align Objects" align:#left checked:false tooltip:"Align the selected objects to either the current parent, new parent or target object as chosen"
		)
		group ""
		(
			button btn_undoBake "-------  UNDO BAKE  -------" width:190 height:20 align:#center tooltip:"The bake process will first HOLD the scene file. Undo Bake will FETCH the scene file back"
		)

		fn f_bakeObjectAdvanced obj copyObj currentParentObj =
		(
			local bakeToObj = copyObj
			if chkbx_currentParent.checked do
			(
				bakeToObj = currentParentObj
			)
			if g_LBTools_BO.f_validateObject targetObj AND chkbx_currentParent.checked == false do
			(
				bakeToObj = targetObj
			)
			with animate on
			(
				for t = g_LBTools_BO.animStart to g_LBTools_BO.animEnd by g_LBTools_BO.stepSize do
				(
					at time t (obj.transform = bakeToObj.transform)
				)
			)
		)
		
		on btn_selectParent pressed do
		(
			if g_LBTools_BO.f_validateObject newParentObj then
			(
				select newParentObj
			)
			else
			(
				messageBox "Error: The NEW PARENT object is not valid.\n\nPlease check to see if it exists and pick it again."
				g_LBTools_BO.f_openBakeObjectsUI()
			)
		)
		
		on btn_selectTarget pressed do
		(
			if g_LBTools_BO.f_validateObject targetObj then
			(
				select targetObj
			)
			else
			(
				messageBox "Error: The TARGET object is not valid.\n\nPlease check to see if it exists and pick it again."
				g_LBTools_BO.f_openBakeObjectsUI()
			)
		)
		
		on btn_clearParent pressed do
		(
			newParentObj = undefined
			pbtn_parent.object = undefined
			chkbx_unlink.checked = true
			if g_LBTools_BO.f_validateObject targetObj do
			(
				chkbx_unlink.enabled = true
			)
			if rdobtns_align.state == 2 do
			(
				rdobtns_align.state = 1
			)
		)
		
		on pbtn_parent picked pickedObj do
		(
			newParentObj = pickedObj
			chkbx_unlink.enabled = false
		)
		
		on btn_clearTarget pressed do
		(
			targetObj = undefined
			pbtn_target.object = undefined
			pbtn_parent.enabled = true
			btn_clearParent.enabled = true
			chkbx_unlink.enabled = false
			chkbx_unlink.checked = true
			chkbx_linkToTarget.enabled = false
			chkbx_linkToTarget.checked = false
			if rdobtns_align.state == 3 do
			(
				rdobtns_align.state = 1
			)
		)
		
		on pbtn_target picked obj do
		(
			targetObj = obj
			chkbx_linkToTarget.enabled = true
			chkbx_unlink.enabled = true
			if g_LBTools_BO.f_validateObject newParentObj do
			(
				chkbx_unlink.enabled = false
			)
		)
		
		on chkbx_currentParent changed theState do
		(
			if theState then
			(
				pbtn_target.enabled = false
				chkbx_linkToTarget.enabled = false
				chkbx_linkToTarget.checked = false
				chkbx_unlink.enabled = false
				chkbx_unlink.checked = true
				btn_clearTarget.enabled = false
			)
			else
			(
				pbtn_target.enabled = true
				btn_clearTarget.enabled = true
				if g_LBTools_BO.f_validateObject newParentObj do
				(
					chkbx_unlink.enabled = false
					chkbx_unlink.checked = false					
				)
				if g_LBTools_BO.f_validateObject targetObj do
				(
					chkbx_linkToTarget.enabled = true
					chkbx_unlink.enabled = true
					chkbx_unlink.checked = true
				)
			)
		)
		
		on chkbx_linkToTarget changed theState do
		(
			if theState then
			(
				chkbx_unlink.enabled = false
				pbtn_parent.enabled = false
				btn_clearParent.enabled = false
			)
			else
			(
				if g_LBTools_BO.f_validateObject newParentObj == false do
				(
					chkbx_unlink.enabled = true
				)
				pbtn_parent.enabled = true
				btn_clearParent.enabled = true
			)
		)
		
		on chkbx_unlink changed theState do
		(
			if theState do
			(
				chkbx_linkToTarget.checked = false
			)
		)

		on chkbx_align changed theState do
		(
			if theState then
			(
				rdobtns_align.enabled = true
			)
			else
			(
				rdobtns_align.enabled = false
			)
		)
		
		on rdobtns_align changed state do
		(
			case rdobtns_align.state of
			(
				2: (
						if (g_LBTools_BO.f_validateObject newParentObj) == false do
						(
							messageBox "Error: The NEW PARENT object is not valid.\n\nPlease check to see if it exists and pick it again."
							pbtn_parent.object = newParentObj = undefined
							rdobtns_align.state = 1
							chkbx_align.checked = false
							rdobtns_align.enabled = false
						)
					)
				3: (
						if (g_LBTools_BO.f_validateObject targetObj) == false do
						(
							messageBox "Error: The TARGET object is not valid.\n\nPlease check to see if it exists and pick it again."
							pbtn_target.object = targetObj = undefined
							rdobtns_align.state = 1
							chkbx_align.checked = false
							rdobtns_align.enabled = false
						)
					)
			)
		)
		
		on btn_undoBake pressed do
		(
			if (queryBox "Are you sure you want to UNDO BAKE?" title:"") do
			(
				fetchMaxFile quiet:true
				g_LBTools_BO.f_openBakeObjectsUI()
			)
		)

		on btn_bakeAdvanced pressed do
		(
			if selection.count > 0 then
			(
				holdMaxFile()
				setCommandPanelTaskMode #create
				local a_selObjs = selection as array
				local errorFound = false
				for obj in a_selObjs while not keyboard.escPressed do
				(
					local currentParentObj = obj.parent
					if chkbx_currentParent.checked AND currentParentObj == undefined do
					(
						local msgtxt = stringStream ""
						format "Bake Objects Error: object < % > has no parent\n" obj.name
						format "Error: object < % > has no parent\n" obj.name to:msgtxt
						messageBox msgtxt title:""
						errorFound = true
					)
					if errorFound == false do
					(
						local copyObj = copy obj
						g_LBTools_BO.f_unlockAllTracks obj
						g_LBTools_BO.f_removeBindings obj
						
						if chkbx_unlink.checked then
						(
							obj.parent = undefined
						)

						if chkbx_linkToTarget.checked AND g_LBTools_BO.f_validateObject targetObj do
						(
							obj.parent = targetObj
						)
						
						f_bakeObjectAdvanced obj copyObj currentParentObj
				
						if chkbx_align.checked == false then
						(
							obj.transform = copyObj.transform
						)
						else
						(
							case rdobtns_align.state of
							(
								1: (if g_LBTools_BO.f_validateObject currentParentObj do (obj.transform = currentParentObj.transform))
								2: (obj.transform = newParentObj.transform)
								3: (obj.transform = targetObj.transform)
							)
						)
						
						if chkbx_unlink.checked == false AND chkbx_linkToTarget.checked == false AND g_LBTools_BO.f_validateObject targetObj AND g_LBTools_BO.f_validateObject currentParentObj do
						(
							obj.parent = currentParentObj
						)
						if g_LBTools_BO.f_validateObject newParentObj do
						(
							obj.parent = newParentObj
						)
						delete copyObj
					)
				)
			)
			else
			(
				messageBox "Error: Please select one or more objects first." title:""
			)
		)
	) --// end rollout

	rollout aboutScript_rollout "About"
	(
		label lbl_author "Author - Lorne Brooks"
		hyperLink myEmail "lbtools2k@gmail.com" color:(color 0 0 255) hovercolor:(color 0 0 190) visitedcolor:(color 190 0 0) align:#center address:"mailto:lbtools2k@gmail.com"
		label lbl_version "Version 1.4 - 3 May 2020"
	) --// end rollout
	
	fn f_initializeINI =
	(
		if NOT (doesFileExist g_LBTools_BO.bakeObjectsINI) then
		(
			setIniSetting g_LBTools_BO.bakeObjectsINI "Bake Objects" "Position" ((mouse.screenpos + [-75,-15]) as string)
			g_LBTools_BO.dPos = mouse.screenpos + [-75,-15]
		)
		else
		(
			g_LBTools_BO.dPos = execute(getIniSetting g_LBTools_BO.bakeObjectsINI "Bake Objects" "Position")
		)
	)
	
	--// Load UI
	fn f_openBakeObjectsUI =
	(
		try(closeRolloutFloater g_LBTools_BO.bakeObjectsFloater)catch()
		g_LBTools_BO.bakeObjectsFloater = newRolloutFloater "Bake Objects" 220 440
		g_LBTools_BO.bakeObjectsFloater.pos = g_LBTools_BO.dPos
		addRollout g_LBTools_BO.timeRange_Rollout g_LBTools_BO.bakeObjectsFloater rolledUp:false
		addRollout g_LBTools_BO.bakeObjectsSimple_Rollout g_LBTools_BO.bakeObjectsFloater rolledUp:false
		addRollout g_LBTools_BO.bakeObjects_Rollout g_LBTools_BO.bakeObjectsFloater rolledUp:true
		addRollout g_LBTools_BO.aboutScript_rollout g_LBTools_BO.bakeObjectsFloater rolledUp:true
	)
	
	f_initializeINI()
	f_openBakeObjectsUI()
)

) --// end struct

g_LBTools_BO = s_LBTools_bakeObjects()


/*********************************************************************************************************************************************************************/
--// MACRO CALLS FOR UI
/*********************************************************************************************************************************************************************/

macroscript bakeObjects
	category:"LB Tools"
	buttontext:"{ BAKE-OBJ }"
	icon:#("LBTools_BakeObjects",1)
	tooltip:"Bake Objects 1.4"
(
	g_LBTools_BO.f_bakeObjects()
)

/*********************************************************************************************************************************************************************/
